/* ====================================================================
 *  Copyright 2003-2004 Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

#include "apr.h"
#include "apr_strings.h"
#include "apr_lib.h"
#include "apr_uri.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"

// DBI Pooling Module
#include "mod_dbi_pool.h"

// libdbi Headers
#include <dbi/dbi.h>

// include php support.
#define HAVE_PHP_SUPPORT
// TODO: use configure.in for this

#ifdef HAVE_PHP_SUPPORT
#include <stdlib.h>
#include "zend.h"
#include "zend_qsort.h"
#include "zend_API.h"
#include "zend_ini.h"
#include "zend_alloc.h"
#include "zend_operators.h"
#endif

typedef struct {
    int disabled;
    conn_id *id;
    char *query;
} vhost_dbi_config_rec;

static void *vhost_dbi_create_server_config(apr_pool_t * p, server_rec * s)
{
    vhost_dbi_config_rec *vdbir =
        (vhost_dbi_config_rec *) apr_pcalloc(p, sizeof(vhost_dbi_config_rec));

    vdbir->disabled = 0;
    vdbir->id = NULL;

    return vdbir;
}

#define HOSTNAME_MAGIC_CHAR '.'

/**
 * "foo.home.com" -> "'foo.home.com','*.home.com','*.com','*'"
 *
 */
const char *expand_hostname(request_rec * r, dbi_conn * dbi,
    const char *hostname)
{
    char tmp[MAX_STRING_LEN];
    const char *s, *e;
    char *p;

    tmp[0] = '\0';
    int written = 0;

    if (!(s = ap_strchr_c(hostname, HOSTNAME_MAGIC_CHAR))) {
        p = strdup(hostname);
        dbi_driver_quote_string(dbi_conn_get_driver(dbi), &p);
        strcat(tmp, p);
        free(p);
        return apr_pstrdup(r->pool, tmp);
    }

    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, r->pool,
        "[mod_vhost_dbi] Hostname to expand: %s", hostname);

    do {
        written += (s - hostname);
        //  MAX_STRING_LEN-1 so we still have room for a null at the end
        if (written >= MAX_STRING_LEN - 1) {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r->pool,
                "[mod_vhost_dbi] Populated Hostname String would exceed %d bytes",
                MAX_STRING_LEN);
            return NULL;
        }
        strncat(tmp, hostname, s - hostname);

    } while ((s = ap_strchr_c(hostname, HOSTNAME_MAGIC_CHAR)));

    strcat(tmp, querystring);
    written += strlen(querystring);
    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, r->pool,
        "[mod_authn_dbi.c] Populated result: \"%s\" / %d chars written",
        apr_pstrdup(r->pool, tmp), written);

    return apr_pstrdup(r->pool, tmp);
}

#define QUERYSTRING_MAGIC_CHAR '&'
#define QUERYSTRING_LEFT_DELIM_CHAR '{'
#define QUERYSTRING_RIGHT_DELIM_CHAR '}'

#define EMPTY_VAR ""            /* do NOT set this to NULL! */

const char *populate_querystring(request_rec * r, dbi_conn * dbi,
    const char *querystring)
{
    char tmp[MAX_STRING_LEN];   /* 8 KByte should be enough for everyone :) */
    const char *s, *e;
    char *p;

    tmp[0] = '\0';
    int written = 0;

    if (!(s = ap_strchr_c(querystring, QUERYSTRING_MAGIC_CHAR)))
        return querystring;

    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, r->pool,
        "[mod_vhost_dbi] Querystring to populate: %s", querystring);
    do {
        written += (s - querystring);
        //  MAX_STRING_LEN-1 so we still have room for a null at the end
        if (written >= MAX_STRING_LEN - 1) {
            ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r->pool,
                "[mod_vhost_dbi] Populated string would exceed %d bytes",
                MAX_STRING_LEN);
            return NULL;
        }
        strncat(tmp, querystring, s - querystring);

        if ((s[1] == QUERYSTRING_LEFT_DELIM_CHAR)
            && (e = ap_strchr_c(s, QUERYSTRING_RIGHT_DELIM_CHAR))) {
            const char *e2 = e;
            char *var;

            p = NULL;
            querystring = e + 1;
            e = NULL;
            var = apr_pstrndup(r->pool, s + 2, e2 - (s + 2));

            if (!strcasecmp(var, "RequestHostname")) {
                e = (r->hostname ? r->hostname : EMPTY_VAR);
            }
            else if (!strcasecmp(var, "RemoteIP")) {
                e = (r->connection->remote_ip ? r->connection->
                    remote_ip : EMPTY_VAR);
            }

            /*
             * Everything but the variable values representing fieldnames and tables gets
             * * escaped according to the selected driver 
             */

            if (e != NULL) {
                p = strdup(e);
                dbi_driver_quote_string(dbi_conn_get_driver(dbi_res->conn),
                    &p);
            }

            if (!strcasecmp(var, "ExpandHostname")) {
                e = (r->hostname ? expand_hostname(r, dbi,
                        r->hostname) : EMPTY_VAR);
            }

            if (e == NULL) {
                ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, r->pool,
                    "[mod_vhost_dbi] Unknown variable: %s", var);
                return NULL;
            }

            if (p == NULL) {
                p = strdup(e);
            }
            written += strlen(p);
            if (written >= MAX_STRING_LEN - 1) {
                ap_log_perror(APLOG_MARK, APLOG_ERR, 0, r->pool,
                    "[mod_vhost_dbi] Populated string would exceed %d bytes",
                    MAX_STRING_LEN);
                free(p);
                return NULL;
            }
            strcat(tmp, p);
            free(p);

        }
        else {
            ap_log_perror(APLOG_MARK, APLOG_WARNING, 0, r->pool,
                "[mod_vhost_dbi] Invalid querystring");
            return NULL;

        };

    } while ((s = ap_strchr_c(querystring, QUERYSTRING_MAGIC_CHAR)));

    strcat(tmp, querystring);
    written += strlen(querystring);
    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, r->pool,
        "[mod_authn_dbi.c] Populated result: \"%s\" / %d chars written",
        apr_pstrdup(r->pool, tmp), written);

    return apr_pstrdup(r->pool, tmp);
}

static const char *set_switch(cmd_parms * cmd, void *config, const char *value)
{
    apr_ssize_t pos = (apr_ssize_t) cmd->info;

}

static const command_rec vhost_dbi_commands[] = {
    AP_INIT_TAKE1("VhostDbiConnName",
        set_switch,
        (void *)0,
        RSRC_CONF,
        "name of the dbi_pool connection to use"),
    AP_INIT_TAKE1("VhostDbiQuery",
        set_switch,
        (void *)0,
        RSRC_CONF,
        "name of the dbi_pool connection to use"),
    AP_INIT_FLAG("VhostDbiEnabled",
        set_switch,
        (void *)1,
        RSRC_CONF,
        "turn mod_vhost_dbi on or off."),
    NULL
};



static char *find_path(request_rec * r, const char *host, char **env,
    vhost_dbi_config_rec * vdbir)
{
    dbi_result *res;
    dbi_conn *dbi = NULL;
    char *qstr;
    char *rstr;
    char *path;
    int dbi_err;
    const char *dbi_err_str;

    dbi = dbipool_open(vdbir->conn_name);

    if (dbi == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
            "vhost_dbi: find_path: DBI connection was not acquired!");
        return NULL;
    }

    rstr = malloc(strlen(host) + 1);
    strcpy(rstr, host);

    dbi_driver_quote_string(dbi_conn_get_driver(dbi), &rstr);

    // TODO: Custom SQL Query Here
    qstr =
        apr_psprintf(r->pool,
        "SELECT vdbi_info.root_path, vdbi_info.php_path "
        "FROM vdbi_info, vdbi_alias "
        "WHERE vdbi_info.idx = vdbi_alias.owner_idx AND "
        "vdbi_alias.hostname=%s LIMIT 0,1", rstr);

    res = dbi_conn_query(dbi, qstr);

    if (res == NULL) {
        dbi_err = dbi_conn_error(dbi, &dbi_err_str);
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
            "vhost_dbi:  get_path: SQL Error: (%d) %s on query: %s", dbi_err,
            dbi_err_str, qstr);
        free(rstr);
        dbi_result_free(res);
        dbipool_close(vdbir->conn_name, dbi);
        return NULL;
    }

    switch (dbi_result_get_numrows(res)) {
    case 1:
        break;
    case 0:
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
            "vhost_dbi: get_path: no results for %s", host);
        free(rstr);
        dbi_result_free(res);
        dbipool_close(vdbir->conn_name, dbi);
        return NULL;
    default:
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
            "vhost_dbi: get_path: %s has more than 1 server row, failing.",
            host);
        free(rstr);
        dbi_result_free(res);
        dbipool_close(vdbir->conn_name, dbi);
        return NULL;
    }

    if (dbi_result_next_row(res) == 0) {
        dbi_err = dbi_conn_error(dbi, &dbi_err_str);
        ap_log_error(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, 0, r->server,
            "vhost_dbi:  get_path: SQL Result Error: (%d) %s on query: %s",
            dbi_err, dbi_err_str, qstr);
        free(rstr);
        dbi_result_free(res);
        dbipool_close(vdbir->conn_name, dbi);
        return NULL;
    }


    path =
        apr_psprintf(r->pool, "%s", dbi_result_get_string(res, "root_path"));

    free(rstr);
    dbi_result_free(res);
    dbipool_close(vdbir->conn_name, dbi);

    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
        "get_path: %s from SQL: %s", path, qstr);

    return path;
}

static int vhost_dbi_init_handler(apr_pool_t * pconf, apr_pool_t * plog,
    apr_pool_t * ptemp, server_rec * s)
{

    ap_add_version_component(pconf, "mod_vhost_dbi/0.1.0");

    return OK;
}

static int vhost_dbi_translate_name(request_rec * r)
{
    vhost_dbi_config_rec *vdbir =
        (vhost_dbi_config_rec *) ap_get_module_config(r->server->module_config,
        &vhost_dbi_module);
    const char *host;
    char *path;
    char *env = NULL;
    char *ptr;
    int i;

    if (!(vdbir->disabled == 1)) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
            "vhost_dbi_translate_name: declined http://%s%s module is not configured for this server",
            apr_table_get(r->headers_in, "Host"), r->uri);
        return DECLINED;
    }

    if (!(host = apr_table_get(r->headers_in, "Host"))) {
        if (!vdbir->default_host) {
            ap_log_error(APLOG_MARK, APLOG_ALERT, 0, r->server,
                "vhost_dbi_translate_name: no host found (non HTTP/1.1 request, no default set) %s",
                host);
            return DECLINED;
        }
        else {
            host = vdbir->default_host;
        }
    }

    if (ptr = strchr(host, ':')) {
        *ptr = '\0';
    }

    // TODO: Cache this Result
    path = find_path(r, host, &env, vdbir);

    if (!path) {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, r->server,
            "vhost_dbi_translate_name: no host found in database for %s",
            host);
        return DECLINED;
    }

    apr_table_set(r->subprocess_env, "vhost_dbi_HOST", host);
    apr_table_set(r->subprocess_env, "vhost_dbi_PATH", path);

    r->filename = apr_psprintf(r->pool, "%s/public_html%s", path, r->uri);

#ifdef HAVE_PHP_SUPPORT
    zend_alter_ini_entry("safe_mode", 13, "on", strlen("on"), 4, 16);
    zend_alter_ini_entry("open_basedir", 13, path, strlen(path), 4, 16);
#endif

    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server,
        "vhost_dbi_translate_name: translated http://%s%s to file %s from uri %s and path %s",
        host, r->uri, r->filename, r->uri, path);

    return OK;
}

static void register_hooks(apr_pool_t * p)
{
    ap_hook_post_config(vhost_dbi_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_translate_name(vhost_dbi_translate_name, NULL, NULL,
        APR_HOOK_FIRST);
}


AP_DECLARE_DATA module vhost_dbi_module = {
    STANDARD20_MODULE_STUFF,
    NULL,                       /* create per-directory config structure */
    NULL,                       /* merge per-directory config structures */
    vhost_dbi_create_server_config,     /* create per-server config structure */
    vhost_dbi_merge_server_config,      /* merge per-server config structures */
    vhost_dbi_commands,         /* command apr_table_t */
    register_hooks              /* register hooks */
};
